/* example/test program for the clwp library */
/* C.M. Moerman, June 2002 (c) */
/* This is a timing test comparing with/without task switching */
/* updated December 2002: typo in NP_CLWP making speed test useless ...
   Thanks to Alexander Mironenko for pointing this out */

/* this program should be compiled and run/timed in two different versions:
   compile, run and time with normal compiler settings: task switching version
   compile, run and time with label NO_CLWP defined ( -DNO_CLWP ): non-switching
*/

#include "../clwp.h"
#include <stdio.h>
#include <time.h>

#define MAXCOUNT 1000       /* i++ counter limit in procedure/task */
#define MAXLOOP 10000000    /* timing iter's, about 30 sec's on my P-III */

int counter = 0;            /* incremented by either procedure or task */
time_t start_time = 0;      /* used for stroing start time */



/* ------------ the two ways to do the program: --------------- */

#ifdef NO_CLWP              /* main procedure WITHOUT task switching */

#define VERSION "(not switching)"
#define RUN proc(&ids);

  void proc(void *param)    /* this proc will be started as procedure */
  {
    int i, temp;
    for (i=0; i<MAXCOUNT; i++) temp++;        /* do some 'processing' */

    counter++;              /* count, so we can check number of iterations */
  }

#else                       /* same WITH task/thread switching */

#define VERSION "(switching !!!)"
#define RUN clwpYield();

  void proc(void *param)    /* this proc will be started as thread */
  {    int i, temp;

    while(1)                /* running for as many times as the main */
    {
        for (i=0; i<MAXCOUNT; i++) temp++;   /* do some 'processing' */
        counter++;          /* count, so we can check number of iterations */
        
        clwpYield();        /* switch back to main program */
    }
  }

#endif


/* -------------- the main program -------------------- */

int main()
{
    int i;
    int ids;                /* to store thread ID if needed */

    printf("\nCLWP: Example4: time the CLWP task switching.\n");
    start_time = time(NULL); /* record starting time */
    
    if(clwpInit(1))         /* initialise the threading system */
    { 
        ids = clwpSpawn(proc, &ids, 4096, 1, TRUE);

        for(i=0; i<MAXLOOP; i++)
        {
            RUN             /* either by procedure call or task switch */
        }

        clwpKill(ids);
    }
                            /* print time used */
    printf("Ready, %d switches in %d seconds\n",
        counter, (int)difftime(time(NULL), start_time));
    return(0);
}

/* end of file */
