/*
 * Allegro mappy framework for gaming, based on Mappy demo game
 * Stripped by C.M. Moerman, februari 2002
 * intended for tile-based games
 *
 * Version 0.1: Based on Mappy demo file
 * see header file for (limited) user documentation
 *
 * Version 1.0 (April 19, 2002; Kees Moerman)
 * - added logging support
 * Version 1.1 (May 24, 2002; Kees Moerman)
 * - improved display selection, added options FM_8BIT, FM_ACCEL
 * Version 1.2 (March 15, 2003)
 * - Do not use MapLoadVRAM under if no hw accel. (problems in Pinpin)
 * - re-initialise timers in case of redraw
 */


#include <stdlib.h>
#include <stdio.h>
#include <strings.h>
#include <allegro.h>
#include <logging.h>
#include "mappyal.h"                    /* the mappy game library */
#include "fm_lib.h"                     /* allegro/mappy wrapper library */
 
// ifdef ALLEGRO_DOS
// define MapLoadVRAM(x) MapLoad(x)       /* V1.2 problems using Mappy VRAM under DOS? */
// endif

int fm_redraw = 0;			            /* Needed if surfaces are lost */
int fm_hwaccel = 0;		                /* Enable hardware acceleration if available */
BITMAP * display1, * display2;	        /* For hardware page flipping */

BITMAP * parallaxbm = NULL;             /* used for background */

int swidth, sheight, displaymode;       /* globals for display mode and size */
int fm_resolution[10] = { 16, 15, 32, 24, 8, 0 };

char errorbuffer[1024];                 /* collected error messages */
int errorchars = 0;                     /* plus current position */

volatile int fm_framer;                 /* refresh if no vsync: incr @ xx Hz timer*/
volatile int fm_frame;                  /* incremented each actual frame 0..N */
volatile int fm_fps;                    /* copy of frame_count each second */
volatile int fm_seconds;                /* elapsed number of seconds */

/* ---------------- Timer functions ---------------------------- */

void frame_timer()                      /* increment frame_time */
{
    fm_framer++;
}
END_OF_FUNCTION(frame_timer);

void fps_proc()                         /* calc frame rate, inc seconds */
{
    fm_fps = fm_frame;                  /* copy once per second and reset */
    fm_frame = 0;
    fm_seconds++;
}
END_OF_FUNCTION(fps_proc);

void init_mytimers()                    /* here we set up game timers etc */
{   int frame_rate = 75;                /* assume 75 images/second */

    LOCK_VARIABLE(fm_framer);           /* may not be swapped out! */
    LOCK_VARIABLE(fm_frame);
    LOCK_VARIABLE(fm_fps);
    LOCK_VARIABLE(fm_seconds);
    LOCK_FUNCTION(frame_timer);
    LOCK_FUNCTION(fps_proc);
	
    if(displaymode >> FM_RATE) frame_rate = displaymode >> FM_RATE;
    install_int_ex(frame_timer, MSEC_TO_TIMER(1000/frame_rate));
    install_int_ex(fps_proc, MSEC_TO_TIMER(1000));
    fm_framer = 0;
    fm_seconds = 0;
}



/* --------------------- display initialisation ---------------------- */
       
void RenderInitDisplays (void)
{
/* Now draw a border to show that DrawMap clips properly */
    log_text(3, "fm: RenderInitDisplay");
	if (!fm_hwaccel) {
//		rect (backbuffer, 3, 3, swidth-4, sheight-4, makecol (255, 255, 255));
//		rect (backbuffer, 9, 9, swidth-10, sheight-10, makecol (255, 255, 255));
	} else {
		acquire_bitmap (display1);
		clear (display1);
//		rect (display1, 3, 3, swidth-4, sheight-4, makecol (255, 255, 255));
//		rect (display1, 9, 9, swidth-10, sheight-10, makecol (255, 255, 255));
		release_bitmap (display1);
		acquire_bitmap (display2);
		clear (display2);
//		rect (display2, 3, 3, swidth-4, sheight-4, makecol (255, 255, 255));
//		rect (display2, 9, 9, swidth-10, sheight-10, makecol (255, 255, 255));
		release_bitmap (display2);
	}
}

void RedrawNeeded (void)                /* set in case window looses focus */
{                                       /* callback from set_display_switch_callback */
  	fm_redraw = 1;
}

int fm_init(int mode, int width, int height, char *mapfile)
{
    RGB blackrgb = { 0, 0, 0 };
    int prefmode = 0;
    int i = 0;

    int window_mode = GFX_AUTODETECT;
    displaymode = mode;                 /* set global */
        
    swidth = width;                     /* copy to globals */
    sheight = height;

    log_num(2, "fm: init ", displaymode);
    log_text(2, allegro_id);

	allegro_init();	/* Normal Allegro calls to start with */
	install_timer();
	install_keyboard();
	install_mouse();

    if (install_sound(DIGI_AUTODETECT, MIDI_AUTODETECT, 0))
    {
        log_text(1, "fm: Sound init error");
        log_text(1, allegro_error);
        allegro_message("Error initialising sound\n%s\n", allegro_error);
        install_sound (DIGI_AUTODETECT, MIDI_NONE, NULL);
    }
    
	init_mytimers();
    errorchars = 0;
    strcpy(errorbuffer, "");

/* Change set_color_depth to see maps change depth automagically... */
    
#ifdef ALLEGRO_DOS
	prefmode = (displaymode & FM_8BIT) ? 8 : 16;
#else
    prefmode = desktop_color_depth();
    log_num(2, "fm: Desk top depth ", prefmode);
	prefmode = (displaymode & FM_8BIT) ? 8 : 16;
#endif
    set_color_depth (prefmode);         /* for some reason, desktop screws up the colors */

    while( (set_gfx_mode (window_mode, swidth, sheight, 0, 0)<0) && fm_resolution[i] )
    {
        prefmode = fm_resolution[i];    /* scan list of preferred depths */
        set_color_depth(prefmode); i++;
    }
    if( fm_resolution[i] == 0)          /* failed? log reason */
    {   log_text(1, "Can not initialise Allegro");
        sprintf(errorbuffer, "Mode %x, %d x %d", window_mode, swidth, sheight);
        log_text(1, errorbuffer);
        log_text(1, allegro_error);
            allegro_message ("Size %d*%d unavailable!\n", swidth, sheight);
        return 0;
    }

#ifdef ALLEGRO_WINDOWS
    set_display_switch_mode(SWITCH_AMNESIA);    /* pause, loose display contents */
  	set_display_switch_callback (SWITCH_IN, RedrawNeeded);  /* getting back focus? */
#endif
	
    display2 = NULL;
	if ((displaymode & FM_ACCEL)
	        && gfx_capabilities&GFX_HW_VRAM_BLIT && gfx_capabilities&GFX_HW_VRAM_BLIT_MASKED)
    {
        log_text(2, "fm: try accel");
		display1 = create_video_bitmap (swidth, sheight);
		display2 = create_video_bitmap (swidth, sheight);
		
		if (display1!=NULL && display2!=NULL) /* OK, bitmaps allocated! */
		{
		    backbuffer = display2;
		    show_video_bitmap (display1);
		    if (bitmap_color_depth(screen)==8) set_palette (desktop_palette);
		
/* Load the FMP file, exit if error */
		    if ((displaymode & FM_MAPMEM) ? MapDecodeVRAM(mapfile) : MapLoadVRAM(mapfile) )
		    {   log_num(1, "fm: Error loading map ", maperror); allegro_exit(); return 0; }
		
		    fm_hwaccel = 1;
            log_text(2, "fm: Accelerated");
		}
		else if(display1 != NULL) { destroy_bitmap( display1); }
	}
	
	if(display2==NULL)                  /* did not succeed with hw accell */
	{
		backbuffer = create_bitmap (swidth, sheight);	/* Double buffer */
		if (backbuffer==NULL) { log_text(1, "No buffer"); MapFreeMem (); exit (0); }
		clear (screen); clear (backbuffer);
		if (bitmap_color_depth(screen)==8) set_palette (desktop_palette);

/* Load the FMP file, exit if error */
		if ((displaymode & FM_MAPMEM) ? MapDecode(mapfile) : MapLoad(mapfile) )
		{   log_num(1, "fm: Error loading map ", maperror); allegro_exit(); return 0; }

		fm_hwaccel = 0;
        log_text(2, "fm: Not accelerated");
	}
	if (bitmap_color_depth(screen)==8) set_color (0, &blackrgb);

/* The next function is new, you don't have to call it, but it may
 * speed up some things and allows you to use maparraypt, eg:
 * if (maparraypt[y][x]==0) blah blah; see readmeal.txt
 */
   MapGenerateYLookup (); /* This can speed up things like collision checking */

/* Now all graphics are loaded, I can set colour 0 (the border) to
 * black if we are in 8bit
 */
	if (bitmap_color_depth(screen)==8)
    { MapSetPal8 (); set_color (0, &blackrgb); }
 
	MapInitAnims ();					/* Reset animations */
    log_num(2, "fm: initialised ", prefmode);
    return 1;
}

void fm_setbackground(BITMAP *backimage)
{
    if(backimage == NULL)
    { log_text(1, "fm: No BG"); exit (0); }
	parallaxbm = MapMakeParallaxBitmap (backimage, fm_hwaccel);
    if(parallaxbm == NULL)
    { log_num(1, "fm: Error setting BG", maperror);
      if(fm_hwaccel)
      { parallaxbm = MapMakeParallaxBitmap (backimage, 0);
        if(parallaxbm == NULL)
        { log_num(1, "fm: Error retry BG", maperror); exit (0); }
      }
    }
	RenderInitDisplays();
    log_text(2, "fm: BG set");
}

void fm_updateparallax(int mapxoff, int mapyoff, int minx, int miny, int maxx, int maxy)
{
    if (parallaxbm != NULL)
		MapDrawParallax (backbuffer, parallaxbm, mapxoff, mapyoff,minx, miny, maxx, maxy);
}



/* ------------------------------ Display update ----------------------------- */

void fm_update_display(void)
{
    int frame_rate = 75;
    
    if(displaymode & FM_ISYNC)          /* wait for sync if needed */
    {   while( !fm_framer && !key[KEY_ESC])
            { };                        /* wait till IRQ sets flag */
        fm_framer = 0;
    }
    
	if (!fm_hwaccel)
	{   /* Sync in with the screen, then copy the doublebuffer to display */
		if (displaymode & FM_VSYNC) vsync ();
//        acquire_bitmap(backbuffer);
//        acquire_bitmap(screen);
		blit (backbuffer, screen, 0, 0, 0, 0, swidth, sheight);
	}
	else
	{   /* Auto-Sync with the screen, then swap display bitmaps (hardware) */
//        acquire_bitmap(backbuffer);
//        acquire_bitmap(screen);
		if (backbuffer==display1)
		{	show_video_bitmap( display1); backbuffer = display2; }
		else
		{	show_video_bitmap( display2); backbuffer = display1; }
	}

	if (fm_redraw > 0) {
		RenderInitDisplays ();
		MapRestore ();
		fm_redraw = -1;
		
        if(displaymode >> FM_RATE) frame_rate = displaymode >> FM_RATE;
        install_int_ex(frame_timer, MSEC_TO_TIMER(1000/frame_rate));
        install_int_ex(fps_proc, MSEC_TO_TIMER(1000));
	}
    fm_frame++;                         /* and again a frame processed */
	
//    release_bitmap(screen);
//    release_bitmap(backbuffer);
}



/* --------------------------- Exit and error messages --------------- */

void fm_exit(void)                      /* clean up and back to text */
{
    log_text(2, "fm: shut down");
	if (!fm_hwaccel) destroy_bitmap( backbuffer);
    else { destroy_bitmap( display1); destroy_bitmap( display2); }
    if(parallaxbm != NULL) destroy_bitmap( parallaxbm);
	MapFreeMem ();
	allegro_exit();
    log_text(2, "fm: done");
}

void fm_error(char *errmess)            /* exit due to error */
{
    strcpy(&errorbuffer[errorchars], errmess);
    errorchars += strlen(errmess);
    log_text(1, "fm: fm_error called");
    log_text(1, errorbuffer);
    allegro_message("Fatal error: %s", errorbuffer);
    fm_exit();
    exit(1);
}

void fm_warning(char *errmess)          /* print/wait on message */
{
    strcpy(&errorbuffer[errorchars], errmess);
    errorchars += strlen(errmess);
    log_text(1, "fm: fm_warning called");
    log_text(1, errorbuffer);
    allegro_message("Warning: %s\n", errorbuffer); 
}

/* end of include file */

