/*
    File logging utilities for error checking
    C.M. Moerman, April 19, 2002 (c)
    Updated 19 jan 2003: splint error messages removed (*@ .... @*)
        run splint with +posixlib option!

    Interface:

        void log_open(unsigned int level, char *logfile);
        void log_text(unsigned int level, char *message);
        void log_close(void);
        void log_num(unsigned int level, char *message, unsigned int number);
        
        (level indicates level of detail to log)

        For explanation see logging.h
*/

#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <time.h>
#include <strings.h>



/* local variable declarations */

static int log_handle = -1;             /* file handle of log file */
static time_t start_time = 0;           /* used for current run time */
static char int_buffer[10];             /* to fastconvert int to ascii */
static unsigned int log_level;          /* level of detail to log */
#define LOGNUMLEN 6


/* actual code incl helper routines */

/* splint complains due to use of global int_buffer */
/*@ -statictrans @*/ /*@ -globstate @*/  /*@ -mustfreefresh @*/    

static char *inttoa(unsigned int input) /*@ +charintliteral @*/     
{
    int_buffer[0] = '0' + ((input/10000) % 10);
    int_buffer[1] = '0' + ((input/1000) % 10);
    int_buffer[2] = '0' + ((input/100) % 10);
    int_buffer[3] = '0' + ((input/10) % 10);
    int_buffer[4] = '0' + (input % 10);
    int_buffer[5] = ' ';                /* trailing space */
    int_buffer[6] = '\0';               /* terminating zero */
    return int_buffer;
}
void log_open(unsigned int level, char *logfile)
{                                       /* unbuffered I/O, flushed */
    char *ptime;
    log_handle = open( logfile, O_RDWR | O_CREAT | O_TRUNC, 0777);
//  fm_assert(log_handle != -1);
    if(log_handle == -1) return;
    start_time = time(NULL);
    ptime = ctime(&start_time);         /* current time as string */
    (void) write( log_handle, ptime, strlen(ptime));
    (void) write( log_handle, "Start log, level ", 17);
    (void) write( log_handle, inttoa(level), LOGNUMLEN);
    (void) write( log_handle, "\n", 1);
    log_level = level;                  /* save level as global */
}

void log_text(unsigned int level, char *message)
{
    if(log_handle == -1) return;        /* logging switched on? */
    if(level > log_level) return;       /* don't log if too detailed */
    (void) inttoa( (unsigned int)difftime(time(NULL), start_time));
    (void) write( log_handle, int_buffer, LOGNUMLEN);
    (void) write( log_handle, message, strlen(message));
    (void) write( log_handle, "\n", 1);
}

void log_num(unsigned int level, char *message, unsigned int number)
{
    if(log_handle == -1) return;        /* logging switched on? */
    if(level > log_level) return;       /* don't log if too detailed */
    (void) inttoa( (unsigned int)difftime(time(NULL), start_time));
    (void) write( log_handle, int_buffer, LOGNUMLEN);
    (void) write( log_handle, message, strlen(message));
    (void) write( log_handle, inttoa(number), LOGNUMLEN-1);
    (void) write( log_handle, "\n", 1);
}

void log_close(void)
{
    if(log_handle == -1) return;        /* logging switched on? */
    (void) write( log_handle, "Stop log\n", 9);
    (void) close( log_handle);
}

/* end of library file */

