/*
    File logging utilities for error checking
    C.M. Moerman, April 19, 2002 (c)

    If the log file is not opened, no text will be written, and effectively
    logging is turned of. So you do not have to put any test before any
    other logging function if you don't want logging, just don't open.

    Logging is only done for those messages where the first argument
    (level) is smaller or equal to the level passed during 'log_open'.
    I.e. when '2' is passed to 'log_open', only log messages with
    level 0, 1 and 2 are actually logged. This way, the detail of
    the logging can be influenced.
    
    Each message is logged with a time stamp (seconds since program start),
    time stamp is 5 digits.
    
    Existing log files will be overwritten silently.
    
    Example syntax:

        if(OPTION('d')) log_open(OPTION('d'), "pinpin.log");
        log_text(1, "Start app");       // logged always
        log_num(2, "Hiscore", score);   // logged only if -d option >= 2
        log_close();
    
    Interface:
*/

/* open the log file, specify log detail level (larger is more detail) */
    void log_open(unsigned int level, char *logfile);

/* log a text message as "<timestamp> <message>" */
    void log_text(unsigned int level, char *message);

/* log a text message with additional numeric info (max 5 digits) */
    void log_num(unsigned int level, char *message, unsigned int number);

/* close the log file */
    void log_close(void);

/* end of header file */

