/*
 * Pinguin game made with Allegro and Mappy
 * C.M. Moerman, Februari-September 2002
 * Version 0.1: Based on Mappy demo file
 * Version 0.2: Rewritten, actual movement
 * Version 0.3: Restructured, help texts added
 * Version 0.4: Baby pinguins and scoring added
 * Version 0.5: (24feb02) air bubbles, upward flow
 * Version 0.6 march 2002: quit handling, intro/exit screens, snow
 * Version 0.7: april 2002: fish
 * Version 1.0: beta release e.g. to Wout
 * Version 1.1: bug fix in Mappy (release 1.1c)
 * Version 1.2: switch to 8 bit, DOS version, -s/-v
 * Version 1.3: new mappy, ???? (ongoing)
 *              July 1, 2002 multi-threading via CLWP library
 *              Swim back of babies using CLWP
 *              July 15, 2002 multi-language support
 * Version 1.4: August 24: read .MAR level files from grab file
 *                 (still to load basic file from grab)
 *              Sept 2: volume bars; bug in baby swimback
 *              Sept 5: enter name in high-score table
 * Version 1.5: Februari 22 new Allegro library version 4.1.8
 *              Compiles with #define ALLEGRO_NO_COMPATIBILITY 1
 *              Quality of service: disable snow, fish if too slow
 *              Better palette handling in 8-bit mode
 *              Display list handling
 *              Crab/Dragon added
 *              New CLWP library version 1.3
 *              Back to mainstream Allegro: 4.0.3 (palette issues in 4.1.8)
 *              Mappy 11D: still problems with LoadVRAM -> fm_lib adapted
 *
 * Command line options:
 *    -w        window instead of full-screen
 *    -l<n>     starting level
 *    -f<filename> separate level/database file
 *    -d<n>     debug info level
 *    -h        hires: work in 16-bit instead of 8 bit (faster in 16-bit windowed)
 *    -s        disable snow
 *    -v        disable fish
 *    -e        enable experimental features (e.g. waves)
 *
 * To Do:
 * - more graphic card support
 * - should we do a yield_timeslice() on high-perf systems?
 * - Alt-Tabbing seems to give some problems, probably due to a timer halt
 */
 
#define VERSION "1.5" 

#define ALLEGRO_NO_COMPATIBILITY 1
#include "main.h"          /* some globals for use in other files etc */
#include "palette.inc"     /* my_default_palette from paint shop pro */

int level_mapping[128] = { 0, 1, 2, 3, 4, 5, 6, 7, 8, 9};

POINT mapoff = {0, 0};
ANIMAL player = {100.0, 100.0, 0.0, 0.0, 100.0, 100.0, 0.0, 0, FM_FRONT1};

/* definition of babies: */

#define MAXBABY 5                       /* max number of FOLLOWING babies */
BABY babies[MAXBABY];                   /* the trail of babies */
int curbabies = 0;
int total_babies = 6;
int baby_id = -1;                       /* will be block number for baby restore */

#define TRAILSIZE (1<<6)                /* must be power of 2!! */
BABY trail[TRAILSIZE];                  /* for following parent */
int header = 0;                         /* queue head in trail */


/* some general variables */

int quit = 0;                           /* set by Esc key */
int level = 0;                          /* index*2 in layers map */
int airsupply = maxair;                 /* initial supply of air */
int saved_babies = 0;
int won = 0;                            /* did we win???? */
int baby_gone = 0;
int fish_on = 1;                        /* do we display the fish? */
int snow_on = 1;                        /* do we display the snow? */

#define sqr(x) ((x)*(x))                /* square value */

void *animal_list = NULL;               /* display list for animals */



/* some primitives */

void MapSelectScenery(void)             /* select scenery layer in map structure */
{
     int map_change_layer_stat = MapChangeLayer(0);
     fm_assert( map_change_layer_stat != -1 );
}

void MapSelectObjects(void)             /* select object layer in map structure */
{
     int map_change_layer_stat = MapChangeLayer(1);
     fm_assert( map_change_layer_stat != -1 );
}

void init_level_mapping()               /* maps might need to be in different */
{                                       /* order as in data file, to ease insertion */
     int i = 0;                         /* of new intermediate levels in later stages */
     char *mappings;                    /* Is specified in help file, at line @8 */
     mappings = find_in_dat(-1, 8);     /* as alpha series, i.e. #@8 ABCDFE */
     fm_assert( *mappings == 'A' );     /* level 1 may not be remapped */
     while (*mappings > ' ')            /* translate alpha to integer */
     {   level_mapping[i++] = *mappings++ - 'A'; }
     level_mapping[i] = -1;             /* end of levels marker */
     log_num(2, "levels found:", i);
}

char scratch[128];                      /* for generating log messages etc */

/* -------------------------------------------------------------- */
/* baby movement routines: how to move and display baby pin's     */
/* note: babies are processed using a separate thread, so I do not*/
/* have to make complex state-machine like code to go through     */
/* different phases (like following, swimming back, ...)          */
/* -------------------------------------------------------------- */

int find_free_baby(void)                /* get a free baby entry in babies*/
{                                       /* so we can have a new baby following */
    int i;                              /* the parent */
    for(i=0; i<MAXBABY; i++)            /* needed for display purposes */
        if (babies[i].number == 0) return i;
    fm_assert(1 == 0);                  /* there should always be a free field... */
}

void baby_thread(void *param)           /* this proc will be started as thread */
{
    BABY *thisbaby;                     /* link to allocated BABY structure */
    int distance, steps, following, i;
    BLKSTR * blk_collide;               /* current block for IGLO detection */
    
    thisbaby = (BABY *)param;           /* cast for ease of access */
    
    log_num(2, "Start baby thread ", clwpGetpid());
    
    /* we have a new baby following the mother/father pinguin */
/*  thisbaby->origx = player.x+(playersprite->w/2);   set in pinguin event handler */
/*  thisbaby->origy = player.y+(playersprite->h-10);  set in pinguin event handler */
    curbabies++;   
    thisbaby->number = curbabies;       /* the how-many'th are we? */
    thisbaby->distance = curbabies * 16;

    MapSelectObjects();                 /* remove baby from map structure */
    ClearCell (thisbaby->origx, thisbaby->origy);
    MapSelectScenery();                 /* go check for IGLO */
    play_sample(data[FM_BIEPBIEP].dat, 128, (player.x-mapoff.x) * 250 / SWIDTH, 1000, 0);
    score += 100;
    
    following = 1;
    while(following)
    {                                   /* set up for display routine */
#define curtrail trail[ (header - thisbaby->distance) & (TRAILSIZE-1)]
        thisbaby->i_image = curtrail.i_image+2; /* main_pinguin+2 => small one */
        thisbaby->x = curtrail.x; thisbaby->y = curtrail.y; 
        thisbaby->mirror = curtrail.mirror; thisbaby->rotate = curtrail.rotate;
#undef curtrail

        MapSelectObjects();             /* just to be sure: */
        blk_collide = Collision( thisbaby->x+(playersprite->w/2), thisbaby->y+(playersprite->h-10));
        MapSelectScenery();             /* go check for IGLO */
        
        if ( (blk_collide != NULL) && (blk_collide->user5 == ITEM_IGLO))
        {                               /* home !!! */
            play_sample(data[FM_BIEPBIEP].dat, 128, (player.x-mapoff.x) * 250 / SWIDTH, 1000, 0);
            thisbaby->number = following = 0;
            score += 100;               /* how to place baby over there? (layers) */
            curbabies--;
            saved_babies++;
            if(saved_babies == total_babies)
                update_help(level_mapping[level], 9);  /* #?91: Won! */
    
            log_num(2, "Homing baby ", clwpGetpid());
            log_num(2, "Stack used: ", clwpStackUsed(clwpGetpid()));
            return;                     /* we're done .. This should kill the thread */
        }
        
        if( (abs(player.vx)+abs(player.vy)<2) && (thisbaby->distance > thisbaby->number*16))
            thisbaby->distance--;       /* close in on distance */
        fm_assert(thisbaby->distance >= 0);
            
        if(abs(player.vx)+abs(player.vy)> maxspeed/2)
        {   if(thisbaby->distance < (TRAILSIZE-1))
                thisbaby->distance++;   /* oops, going to fass for baby? */
            else                        /* lost track... */
            {
                following = 0;
                curbabies--;
                score -= 110;
            }
        }
        displaylist_add(&animal_list,
            data[thisbaby->i_image].dat, 
            thisbaby->x, thisbaby->y+4, thisbaby->mirror, thisbaby->rotate);
        clwpYield();                    /* don't forget other tasks... */
    }
    
    /* baby lost track of the parent, out of follow loop, swim back 'home'*/
    
    log_num(2, "Lost baby ", clwpGetpid());
    
    distance = (int)sqrt( sqr(thisbaby->x - thisbaby->origx) + sqr(thisbaby->y - thisbaby->origy));
    steps = distance/8;                 /* number of steps to take to starting pos */
    thisbaby->i_image = FM_SWIM1+2,     /* main_pinguin+2 => small one */
    thisbaby->rotate = 0; thisbaby->mirror = (thisbaby->x > thisbaby->origx);
    
    play_sample(data[FM_LOST].dat, 128, (player.x-mapoff.x) * 250 / SWIDTH, 1000, 0);
    if(!baby_gone++)                    /* only display this once per level */
        update_help(-1, 7);             /* #@6 baby gone help */
    
    for(i=steps; i>0; i--)
    {                                   /* mapping to screen not yet correct? */
        thisbaby->x += (thisbaby->origx - thisbaby->x)/i;
        thisbaby->y += (thisbaby->origy - thisbaby->y)/i;
        displaylist_add(&animal_list,
            data[thisbaby->i_image].dat, 
            thisbaby->x, thisbaby->y+4, thisbaby->mirror, thisbaby->rotate);
        clwpYield();                    /* one step at a time, switch to next thread */
    }
    
    MapSelectObjects();                 /* place baby back into map structure */
    SetCell (thisbaby->origx, thisbaby->origy, -baby_id-1);
    MapSelectScenery();
    thisbaby->number = 0;               /* free up structure for re-use */
    
    play_sample(data[FM_LOST].dat, 128, (player.x-mapoff.x) * 250 / SWIDTH, 1000, 0);
    
    log_num(2, "Killing baby thread ", clwpGetpid());
    log_num(2, "Stack used: ", clwpStackUsed(clwpGetpid()));
    return;                             /* we're done .. This should kill the thread */
}


/* -------------------------------------------------------------- */
/* player movement routines: how to move and display pinguin      */
/* -------------------------------------------------------------- */

int jump = 1234;                        /* just a marker value */

void init_pinguins(void)                /* set up datastructures etc */
{
    int i;
    player.x = player.px = ConfigBlock()->user5 * mapblockwidth;
    player.y = player.py = ConfigBlock()->user6 * mapblockheight;
    total_babies = ConfigBlock()->user7;
    
    for(i=0; i<TRAILSIZE; i++)          /* all at current location */
    {   trail[i].x = player.x; trail[i].y = player.y;
        trail[i].rotate = 0;   trail[i].mirror = 0;
        trail[i].i_image = FM_FRONT1;
    }
    for(i=0; i<MAXBABY; i++)            /* will be used for baby threads */
       babies[i].origx = babies[i].origy = babies[i].number
           = babies[i].distance = 0;
    curbabies = 0;                      /* nothing following at the moment */
    saved_babies = 0;
}

void pinguin_walk(void)                 /* basic above-water movements */
{
    player.px = player.x; player.py = player.y; /* remember old position */
    if (key[KEY_RIGHT]) player.vx += speed;
    if (key[KEY_LEFT])  player.vx -= speed;
                                        /* check if we have ground below us */
    if (Collision (player.x+(playersprite->w/2), (player.y+playersprite->h)+2))
    {
        if (key[KEY_UP])  player.vy -= 9;   /* ground: can jump */
        else player.vy = 0;             /* not jumping: zero speed in y*/
        player.vy = max(-10, player.vy);    /* no cumulative jumping... */
    }
    else                                /* no ground below us: help..... */
        player.vy += speed * 0.6;       /* gain speed downwards (positive) */

    player.vx *= slowdown_a;            /* reduce previous speed */
    player.x += player.vx;              /* and update position */
    player.y += player.vy;
    
    if (player.vy < 0)                  /* collide upwards: reverse speed */
    {    if (Collision( player.x+playersprite->w/2, player.y))
         {   player.y = player.py; player.vy = -player.vy/8; }
    }
    else                                /* going down? */
    {    if (Collision( player.x+playersprite->w/2, player.y+playersprite->h))
         {    player.y = player.py; player.vy = -player.vy/8; }
    }
                                        /* image decision and movement*/
    player.i_image = ((abs(player.vx) < 0.2) ? FM_FRONT1 : FM_SIDE1) + ((fm_frame/10)&1);
    player.mirror = (player.vx < 0);
    player.rotate = 0;
}

void pinguin_swim(void)                 /* basic below-water movement */ 
{
    float curspeed;                     /* 'abs vector speed' */

    if(player.py < waterlevel)          /* previous was above? splash */
    {    play_sample(data[FM_PLONS].dat,
             player.vy * 16, (player.x-mapoff.x) * 250 / SWIDTH, 1000, 0);
         player.vy /= 2;                /* water breaks fall, reduce speed */
    }
        
    player.px = player.x; player.py = player.y; /* remember old position */
    jump = 1234;                        /* enable jump again for later */
    
    if (key[KEY_RIGHT]) player.vx += speed + abs(player.vy/6);    /* can go any direction */
    if (key[KEY_LEFT])  player.vx -= speed + abs(player.vy/6);    /* when swimming */
    if (key[KEY_UP])    player.vy -= speed + abs(player.vx/6);
    if (key[KEY_DOWN])  player.vy += speed + abs(player.vx/6);
    curspeed = sqrt((player.vx*player.vx)+(player.vy*player.vy));
    if(curspeed > maxspeed)
    {   player.vx *= maxspeed/curspeed; player.vy *= maxspeed/curspeed; }

    player.vx *= slowdown_w; player.vy *= slowdown_w; /* reduce previous speed */
    player.x += player.vx;              /* and update position */
    player.y += player.vy;
    
    if (player.vy < 0)                  /* collide up: reverse speed */
    {    if (Collision (player.x+playersprite->w/2, player.y))
         {   player.y = player.py; player.vy = -player.vy/3; }
    }
    else                                /* going down? */
    {    if (Collision (player.x+playersprite->w/2, player.y+playersprite->h))
         {    player.y = player.py; player.vy = -player.vy/3; }
    }
                                        /* image is more complex .... */
    player.i_image = (((abs(player.vx) + abs(player.vy)) < 0.2) ? FM_FRONT1 : FM_SWIM1) + ((fm_frame/10)&1);
    player.mirror = (player.vx < 0);    /* rotate in swimming direction */
    player.rotate = (((abs(player.vx) + abs(player.vy)) < 0.2)
        ? 0 : fixatan2(itofix(player.vx), -itofix(player.vy)));
    if((abs(player.vx) + abs(player.vy)) < 4)   /* but not complete if still slow... */
        player.rotate = player.rotate * (abs(player.vx) + abs(player.vy))/4;
}

void pinguin_event(void)
{    
    BLKSTR * blk_collide;
    int temp, temp_free;
    
    blk_collide = Collision(player.x+(playersprite->w/2), player.y+(playersprite->h-10));
    if ( blk_collide != NULL)
    {   
        switch( blk_collide->user5)
        {
            case ITEM_POINTS:
            {   score += blk_collide->user3;
                ClearCell (player.x+(playersprite->w/2), player.y+(playersprite->h-10));
                play_sample(data[FM_BLUB].dat, 128, (player.x-mapoff.x) * 250 / SWIDTH, 1000, 0);
                break;
            }
            case ITEM_HELP:
            {   if(help_active == 0)
                    update_help(level_mapping[level], blk_collide->user3); break;
            }
            case ITEM_AUW:
            {   score -= blk_collide->user3;
                play_sample(data[FM_BLUB].dat, 128, (player.x-mapoff.x) * 250 / SWIDTH, 1000, 0);
                break;
            }
            case ITEM_BABY:
            {   if((curbabies < MAXBABY) && (curbabies < 3))
                {
                    baby_id = blk_collide->user6;   /* keep for later restore */
                    temp_free = find_free_baby();
                    babies[temp_free].origx = player.x+(playersprite->w/2);
                    babies[temp_free].origy = player.y+(playersprite->h-10);
                    temp = clwpSpawn(baby_thread, &babies[temp_free], 16384, 1, TRUE);
                    fm_assert(temp != 0);
                }
                break;
            }
            case ITEM_SLOW:                 /* slow_down due to weed etc */
            {
                if ((abs(player.vx)+abs(player.vy)) > maxspeed/5)
                {    player.vx /= 1.4; player.vy /= 1.4;  }
                break;
            }
            case ITEM_BUBBLE:               /* to do: upwards flow movement */
            {
                if ((abs(player.vx)+abs(player.vy)) < maxspeed-1)
                    player.vy -= 0.5;
                break;
            }
            case ITEM_AIR:                  /* to do: refill air supply */
            {
                airsupply = min(airsupply+10, maxair);
                break;
            }
            case ITEM_IGLO:   { break; }    /* handled by babies themselves */
            default: fm_assert(0);          /* illegal value in item */
        }
    }
}

void update_pinguin(void)               /* main player movement */
{
/* player movement: this is the tricky part. As the human player is
   focussed on this image, movement must be 'high quality', whatever
   that is. */
    if(player.y < waterlevel)           /* above water: keep straight up */
        pinguin_walk();
    else                                /* below water: we go swimming */
        pinguin_swim();
        
    if(player.y-10 < waterlevel)        /* above water: get air */
        airsupply = min(airsupply+maxair/50, maxair);
    else                                /* below water: use air */
        airsupply = max(airsupply-1, 0);
        
/* constrain player within screen boundaries */
    player.x = max(min(player.x, mapwidth*mapblockwidth-32-(LMARGIN+RMARGIN)), 0);
    player.y = max(min(player.y, mapheight*mapblockheight-32-(TMARGIN+BMARGIN)), 0);

    if (player.vx < 0)  /* collide left/right: reverse speed */
    {    if ( /* !key[KEY_F9] && */
              Collision (player.x, player.y+(playersprite->h-10)))
         {   player.x = player.px; player.vx = -player.vx/3;  }
    }
    else
    {    if ( /* !key[KEY_F9] && */
              Collision (player.x+playersprite->w, player.y+(playersprite->h-10)))
         {   player.x = player.px; player.vx = -player.vx/3; }
    }

    if (player.y < 0) player.y = 0;     /* keep within screen...... */

/* Change layer to objects, see if player touches, then change back to layer 0 */
    MapSelectObjects();
    pinguin_event();                    /* check for special events (odd layers) */
    MapSelectScenery();

#define curtrail trail[header & (TRAILSIZE-1)]
    curtrail.x = player.x; curtrail.y = player.y;
    curtrail.rotate = player.rotate; curtrail.mirror = player.mirror;
    curtrail.i_image = player.i_image;
    header++;
#undef curtrail

    displaylist_add(&animal_list,       /* ready to display at frame rate */
        data[player.i_image].dat, player.x, player.y+4, player.mirror, player.rotate);
}

int crab_id1 = 0;
int crab_id2 = 0;

void crab(int x, int y)
{
    int direction = 1;
    int timer = 0;

    while(!Collision (x, y+3+((BITMAP *)data[FM_CRAB1].dat)->h)) y++;
    y--;                                /* let crab sink until on bottom */
     
    while(1)
    {
        x += direction;
        if(direction > 0)
        {   if (Collision( x+(((BITMAP *)data[FM_CRAB1].dat)->w), y+8))
            { direction = -1; x += 2*direction; }
        }
        else
        {   if (Collision(x, y+8)) { direction = 1; x += 2*direction; }
        }
        if(rand_n(20) == 0)             /* close in on player? */
        {   if(x > player.x) direction = -1; else direction = 1;
        }

        if((abs(x - player.x) < 25) && (abs(y-8 - player.y) < 20))
        {                               /* close to player: try to grab */
            displaylist_add(&animal_list,   /* ready to display at frame rate */
                data[FM_CRAB3 + ((timer++ & 0x08)?1:0)].dat, x, y+4, 0, 0);
            if (timer & 1)
            {   score -= 1;
                play_sample(data[FM_CRAB].dat, 128, (player.x-mapoff.x) * 250 / SWIDTH, 1000, 0);
            }
        }
        else                            /* larger distance */
            displaylist_add(&animal_list,   /* ready to display at frame rate */
                data[FM_CRAB1 + ((timer++ & 0x08)?1:0)].dat, x, y+4, 0, 0);
        clwpYield();                    /* don't forget other tasks... */
    }
    return;                             /* end of thread */
}

#define DX_SCALE (1+((score>5500)?1:0)) /* speed up for good players */
#define DY_SCALE 1
#define DRAGONLEN 128                   /* size of tail array: power of two!! */
#define DRAGONSTEP 20                   /* step size in tail array */

void dragon( int x, int y)
{
    int timer = 0, xdirection = 1, ydirection = 1;
    int i;
    int index = 0;
    BABY tail[DRAGONLEN];               /* tail parts location/orientation */
    
    for(i=0; i<DRAGONLEN; i++)          /* initialise task data */
    { tail[i].x = x; tail[i].y = y; tail[i].mirror = 0; tail[i].rotate = 0; }
    
    while(1)                            /* endless task loop */
    {
        while(abs(x - player.x) > SWIDTH) clwpYield(); /* outside display */
        
        if(rand_n(4) != 0) x += DX_SCALE * xdirection;
        if(xdirection > 0)              /* move acc. to direction */
        {   if (Collision( x+(((BITMAP *)data[FM_DRAGON1].dat)->w-8), y+16))
            { xdirection = -xdirection; x += 2 * DX_SCALE * xdirection; }
        }                               /* but turn on walls */
        else
        {   if (Collision( x+8, y+16))
            {   xdirection = -xdirection; x += 2 * DX_SCALE * xdirection; }
        }
        if((rand_n(40) == 0) && (abs(x - player.x) < SWIDTH/2))  /* close in on player? */
        {   if(x > player.x) xdirection = -1; else xdirection = 1;
        }
        
        if(rand_n(2) == 0) y += DY_SCALE * ydirection;  /* up and down: not too far */
        
        if(ydirection > 0)              /* check if not colliding.. */
        {   if (Collision( x+16, y+(((BITMAP *)data[FM_DRAGON1].dat)->h-8)))
            {   ydirection = -ydirection; y += 2 * DY_SCALE * ydirection; }
        }
        else
        {   if (Collision( x+16, y+8))
            {   ydirection = -ydirection; y += 2 * DY_SCALE * ydirection; }
        }
        if(rand_n(40) == 0)             /* close in on player? */
        {   if(y > player.y) ydirection = -1; else ydirection = 1;
        }
        if(y < waterlevel) { y += 6; ydirection = 1; }
        if(y > SHEIGHT)    { y -= 6; ydirection = -1; }
                                        /* play sound only if near */
        if((abs(x - player.x) < 2*SWIDTH/3) && (rand_n(80) == 0))
            play_sample(data[FM_DRAGON].dat, 128,
                max( min( (x-mapoff.x) * 250 / SWIDTH, 255), 0), /* 0:left, 255: right */
                1000, 0);

        displaylist_add(&animal_list,   /* ready to display at frame rate */
            data[FM_DRAGON1 + ((timer++ & 0x08)?1:0)].dat,
                x, y+4, xdirection<0?0:1, 0);

        for(i=0; i<4; i++)
        {
#define CUR_TAIL tail[(index + i * DRAGONSTEP) & (DRAGONLEN-1)]
            displaylist_add(&animal_list,
                data[FM_DRAGON1 + 1 + (i?i:1)].dat,
                CUR_TAIL.x, CUR_TAIL.y, CUR_TAIL.mirror, CUR_TAIL.rotate);
            if((abs(CUR_TAIL.x - player.x) < 20) && (abs(CUR_TAIL.y - player.y) < 15))
            {                           /* close to player: try to grab */
                if (timer & 1)
                {   score -= 4-i;
                    play_sample(data[FM_CRAB].dat, 128, (player.x-mapoff.x) * 250 / SWIDTH, 1000, 0);
                }
            }
        }

        if((abs(x - player.x) < 25) && (abs(y - player.y) < 20))
        {                               /* close to player: try to grab */
            if (timer & 1)
            {   score -= 5;
                play_sample(data[FM_CRAB].dat, 128, (player.x-mapoff.x) * 250 / SWIDTH, 1000, 0);
            }
        }

        i = -1;
        CUR_TAIL.x = x; CUR_TAIL.y = y+4; CUR_TAIL.mirror = xdirection<0?0:1;
        index--;
        clwpYield();                    /* don't forget other tasks... */
    }
}

void crab_thread(void *param)           /* autonomous moving crab/dragon */
{
    int x, y, which;

    y = (*(int *)param) & 0x7F; x = ((*(int *)param)) >> 8 & 0xFF;
    which = (*(int *)param) & 0x80;     /* 0 = crab, 0x80 = dragon */
    sprintf(scratch, "Crab start at %04x: %d, %d", (*(int *)param), x, y);
    log_text(1, scratch);
    x = max(1, min(199, x)); y = max(1, min(18, y));
    x *= mapblockwidth;      y *= mapblockheight; 
    if(which) dragon( x, y);
    else      crab( x, y);
    return;
}

int crab1, crab2;                       /* not local variables! (disappear from scope) */

void kill_crab(void)
{
    if(crab_id1 != 0) clwpKill(crab_id1); /* kill previous instance */
    if(crab_id2 != 0) clwpKill(crab_id2);
    crab_id1 = 0; crab_id2 = 0;
}

void init_crab(void)
{   kill_crab();
    crab1 = ConfigBlock()->user3;       /* get starting information */
    crab2 = ConfigBlock()->user4;
    sprintf(scratch, "Crab test on %d, %d", crab1, crab2);
    log_text(1, scratch);
    if(crab1) crab_id1 = clwpSpawn(crab_thread, &crab1, 16384, 1, TRUE);
    if(crab2) crab_id2 = clwpSpawn(crab_thread, &crab2, 16384, 1, TRUE);
}

void init_level(void)                   /* set up for next level */
{
    log_num(2, "Loading ", level);
    MapDecodeMAR ( data[FM_L1a + 2*level_mapping[level] ].dat, 0); /* load scenery level */
    MapDecodeMAR ( data[FM_L1b + 2*level_mapping[level] ].dat, 1); /* load object level */

    init_pinguins();                    /* set up level */
    init_snow();
    init_crab();
    update_help(level_mapping[level], 0);  /* greating text */
    baby_gone = 0;
    log_num(2, "Loaded ", level_mapping[level]);
}

void check_new_level(void)              /* did we bring them all home? */
{
    if(saved_babies != total_babies) return;
    if(help_active > 0) return;         /* win text still active: wait */
    if(level_mapping[level+1] == -1)    /* max level reached? */
    {
        won = 1; quit = 1;
    }
    else
    {
        level++;
        if(res8bit) fade_out( fade_speed);
        init_level();                   /* set up level */
    }
}


/* -------------------------------------------------------------- */
/* main program loop, main program                                */
/* -------------------------------------------------------------- */

int temp_time = 0;                      /* used to re-set the palette */

void intro_loop(int exit_flag)          /* startup screen, exit screens */
{   int i = 0;                          /* flag: 0 = intro, 1 = exit; */
                                        /* 2 = won all levels, 3 = get name for hiscore */
//    int faded = 0;
    while( keypressed()) { readkey(); } /* empty keyboard input buffer */
    init_snow();
    if( exit_flag != 3)
    {   update_help(-1, 4+exit_flag);   /* #@4, #@5, $@6 greeting, credit, won */
    }
    else
    {   display_hiscore();              /* we have a new hi score */
        read_name_flag = 1;
    }

/* This is the main loop, which is repeated until SPACE or ESC is pressed */
    while (help_active > 0)             /* either time-out or space */
    {
        draw_sprite(backbuffer, data[FM_BACKGROUND].dat, 0, 0);
        
        update_fish();
        draw_fish();
        clwpYield();                    /* other processes need some time? */
        check_clwp_errors();
        quality_of_service();
                                        /* 3 penguins on front page */
        draw_sprite(backbuffer, data[FM_FRONT1+((i>>4)&1)].dat, SWIDTH/2-40, 205);
        draw_sprite(backbuffer, data[FM_SIDE1+(((i+3)>>4)&1)].dat, SWIDTH/2-40+50, 205);
        draw_sprite_h_flip(backbuffer, data[FM_SIDE1+(((i+7)>>4)&1)].dat, SWIDTH/2-40-50, 205);
        
        draw_sprite(backbuffer, data[FM_CARTOON].dat, /* my image */
            SWIDTH-((BITMAP *)data[FM_CARTOON].dat)->w, 0);
            /* SHEIGHT-((BITMAP *)data[FM_CARTOON].dat)->h); */
        
        let_it_snow();
        display_helptext(10, 245);
        display_debug_text();
        i++;                            /* used to animate (counter) */
        
        fm_update_display();
        if(temp_time != fm_seconds)
        { temp_time = fm_seconds; set_palette( data[FM_DEF_PAL].dat); } /* why needed? */
    }
    
    while(key[KEY_SPACE] || key[KEY_ESC]) { }; /* wait for key release */
    if(res8bit) fade_out( fade_speed);
}

int log_counter = 0;                    /* logging counter: log every 512 frames */       

void main_loop(void)
{   
    set_palette( data[FM_DEF_PAL].dat); /* need first to set palette, then */
    playersprite = data[FM_FRONT1].dat; /* for size reference */
    fm_framer = 0;                      /* first time, it may be already incremented */

/* This is the main loop, which is repeated until SPACE or ESC is pressed */
    while (!quit)
    {
        check_new_level();              /* proceed to next level? */

        /* some statistics for logging purposes every 512 image frames (~10 sec) */
        if( !(log_counter++ & 0x1FF)) log_state();
        
/* --- this loop is running at the game rate, not the display frame rate --- */
        while(fm_framer>0)              /* for lower frame rates: updates more often */
        {
/*          keyhandler();               (* menu, function keys etc: in thread! */
            
            fm_framer--;
            if(fm_framer<0) fm_framer = 0;
            if(fm_framer>10) fm_framer = 1; /* break down gracefully .... */
            
            displaylist_clear(&animal_list);
            update_pinguin();           /* move it!! */
            MapUpdateAnims();           /* like bubble animations etc in Mappy */
            if(fish_on) update_fish();
            update_score();
            clwpYield();                /* also give other processes some time */
            check_clwp_errors();
        }
/* --------- end of game loop --------- */

/* set up for display: everything which does needs to be updated,
   but does not have to be done at game rate */
   
        quality_of_service();           /* need to switch off features? */
/* calculate screen coordinates, constrain within boundaries */
        mapoff.x = (player.x+(playersprite->w/2)+10)-(SWIDTH/2);
        mapoff.y = player.y*80/SHEIGHT;
        mapoff.x = max(min(mapoff.x, mapwidth*mapblockwidth-(SWIDTH-(LMARGIN+RMARGIN))), 0);
        mapoff.y = max(min(mapoff.y, mapheight*mapblockheight-(SHEIGHT-(TMARGIN+BMARGIN))), 0);
        
        if(temp_time != fm_seconds)     /* sometimes, the palette is damaged */
        { temp_time = fm_seconds; set_palette( data[FM_DEF_PAL].dat); } /* why needed? */


/* now we are going to display the actual image layers: */        
        acquire_bitmap(backbuffer);

        fm_updateparallax( mapoff.x, mapoff.y,  /* parallax layer (ice bergs) */
            LMARGIN, TMARGIN, SWIDTH-(LMARGIN+RMARGIN), SHEIGHT-(TMARGIN+BMARGIN));
        draw_sprite(backbuffer, data[FM_SUN].dat, SWIDTH*5/8, 40);   /* sun behind all */

        if(fish_on) draw_fish();        /* place fish behind blocks.... */
        
        MapDrawBGT (backbuffer, mapoff.x, mapoff.y,   /* background blocks */
            LMARGIN, TMARGIN, SWIDTH-(LMARGIN+RMARGIN), SHEIGHT-(TMARGIN+BMARGIN));

/* Anything you want BETWEEN back and foreground goes in here... */
        displaylist_draw(animal_list);  /* this does the actual display */
        
/* Change layer to objects, draw, then change back to layer 0 */
        MapSelectObjects();             /* like shell fish items */
        MapDrawFG (backbuffer, mapoff.x, mapoff.y,
            LMARGIN, TMARGIN, SWIDTH-(LMARGIN+RMARGIN), SHEIGHT-(TMARGIN+BMARGIN), 0);
        MapSelectScenery();


/* Lastly, the foreground is drawn */
        MapDrawFG (backbuffer, mapoff.x, mapoff.y,
                LMARGIN, TMARGIN, SWIDTH-(LMARGIN+RMARGIN), SHEIGHT-(TMARGIN+BMARGIN), 0);
        draw_waves();                   /* wave feature */
        if(snow_on) let_it_snow();      /* plus all misc foreground items */
        display_score();                /* like frame with score */
        display_helptext(50, 150);      /* any help messages or overlays */
        display_debug_text();
        
        fm_update_display();
        release_bitmap(backbuffer);
        
        check_clwp_errors();
    }
    
    kill_crab();
    if(res8bit) fade_out( fade_speed);
    while(key[KEY_SPACE] || key[KEY_ESC] || key[KEY_C])
        { };                            /* wait for key release */
}

int res8bit = FM_8BIT;                  /* default 8 bits resolution */
RGB black = {0, 0, 0};                  /* needed in palette handling */
RGB pink = {255, 0, 255};

int main (int argc, char *argv[])
{
/*  command line options:
 *  -w       run in windowed mode
    -l<n>    start at level n
    -f<filename> get level from filename.fmp, database from filename.grb
 */

    char datafilename[256];
    char levelfilename[256];

	Option_comline(argc, argv);			/* process options */
    
//    if(OPTION('d')) log_open(OPTION('d'), "pinpin.log");
    log_open(9, "pinpin.log");
    log_text(1, "Pinpin versie " VERSION ", " __DATE__ " " __TIME__
#ifdef ALLEGRO_DOS
    " under DOS"                        /* DJGPP compiler */
#else
    " under DirectX"                    /* Dev_C++ compiler */
#endif
    );

    strcpy(datafilename, OPTION('f')?OPTIONPTR('f'):"pinpin"); 
    strcpy(levelfilename, datafilename); 
    strcat(datafilename,  ".grb");      /* construct the file names */
    strcat(levelfilename, ".fmp");

    if(OPTION('h')) { res8bit = 0; }
    else { my_default_palette[0] = black; }

    log_text(2, "Open hiscore");
    load_hiscore();

    if(!fm_init( (OPTION('w') ? 0 : FM_FULL) | res8bit
#ifdef ALLEGRO_WINDOWS
               | FM_VSYNC               /* not needed/wanted under DOS? */
#endif
               | (GAME_RATE <<FM_RATE) | FM_ACCEL, /* timer driver game rate */
            SWIDTH, SHEIGHT, levelfilename))  
    {   log_text(1, "Can not initialise Allegro");
        exit(1);
    }
    text_mode(-1);                      /* Allegro text on transparant background */
    log_system_info();                  /* CPU and OS type */
    if(!clwpInit(1))
    {   log_text(1, "Can not initialise CLWP");
        exit(1);
    }
    
    log_text(2, "Set pw");
    packfile_password("moerman");       /* 'encrypted' data file */
    set_palette( my_default_palette);   /* need first to set palette */

    if(!res8bit)
    {
        log_text(2, "Pre-load datafile");
        data = load_datafile(datafilename);
        if(data == NULL) { log_text(1, "data pre-load error"); } /* if not 8 bits: */
        set_palette( data[FM_DEF_PAL].dat); /* need first to set palette, then */
        unload_datafile(data);              /* reload images for correct mapping */
    }

    log_text(2, "Load datafile");
    data = load_datafile(datafilename);
    if(data == NULL) { log_text(1, "data load error"); }
//    fm_hwaccel=0;
    set_palette( data[FM_DEF_PAL].dat); /* final palette from grabber file */
    fm_setbackground(data[FM_BACKGROUND].dat);
    packfile_password(NULL);            /* remove password */

    level = 0;
    if(OPTION('l')) level = OPTION('l')-1; /* for testing other levels */
    if(OPTION('s')) snow_on = 0;
    if(OPTION('v')) fish_on = 0;
     
//  music = load_midi("sounds\\music.mid");
    play_midi(data[FM_MUSIC].dat, 1);

    log_text(2, "init buffers");
    init_buffers();                     /* request various bitmap buffers */
    displaylist_new(&animal_list, 20);  /* new display list for max 20 items */
    init_keyhandler();                  /* initialise key handling thread */
    init_fish();
    init_help();
    init_level_mapping();

    log_text(2, "Intro loop");
    intro_loop(0);                      /* startup display */

    init_level();                       /* player/level initialisation */
    log_text(2, "Main loop");
    main_loop();
    
    if(test_hiscore()) intro_loop(3);
    
    log_text(2, "Save hiscore");
    save_hiscore();

    if(won) intro_loop(2);              /* yes, we're done!!! hi_score etc */    
    log_text(2, "Intro loop exit");
    intro_loop(1);                      /* credit display etc */
    
/* When ESC or SPACE has been pressed, it's time to clean up and leave */
    exit_help();
    exit_buffers();                     /* incl. MapFreeMem(); */
    unload_datafile(data);
    clwpExit();
    fm_exit();
    
    log_close();
    
    return 0;
}

END_OF_MAIN ();

/* end of file */
