/*
 * Pinguin game made with Allegro and Mappy
 * C.M. Moerman, Februari 2002 - Februari 2003
 */

/* main/global definitions for pinguin game */

#define NEW 0              /* get new features in? */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <assert.h>
#include <allegro.h>
#include <options.h>       /* command line option parsing */
#include <mappyal.h>       /* the mappy game library */
#include <fm_lib.h>        /* Allero/Mappy wrapper library */
#include <clwp.h>          /* cooperative multi-threading library */
#include <logging.h>       /* log messages to file */
#include "pingrab.h"       /* grabber-generated header file of data file */

#define GAME_RATE 40       /* animation/game updates/second */

#define speed 1
#define slowdown_w 0.92
#define slowdown_a 0.85
#define waterlevel (SHEIGHT/2+16)
#define maxspeed 15

#define fade_speed 4

#define SWIDTH 640       /* display width of full screen */
#define LMARGIN 16       /* unused margin left */
#define RMARGIN 10       /* unused margin right */
#define SHEIGHT 480      /* idem for height */
#define TMARGIN 28
#define BMARGIN 12
#define AWIDTH (SWIDTH-LMARGIN-RMARGIN)   /* active width */
#define AHEIGHT (SHEIGHT-TMARGIN-BMARGIN)

#define ITEM_POINTS 1 /* find points as stored in user3 */
#define ITEM_HELP   2 /* help sign with text index in user3 */
#define ITEM_AUW    3 /* damage as stored in user3 */
#define ITEM_BABY   4 /* find baby pinguin */
#define ITEM_IGLO   5 /* find iglo (target for babies) */
#define ITEM_BUBBLE 6 /* flow caused by bubble */
#define ITEM_AIR    7 /* air bubble, replenish supply */
#define ITEM_SLOW   8 /* slow down due to sea weed etc */

#define rand_n(n) (rand() % (n))        /* random 0..n-1 */

DATAFILE *data;                         /* Allegro grabber datafile will end up here */

typedef struct POINT { int x; int y; } POINT;

typedef struct ANIMAL {                 /* a pinguin object */
    float x, y, vx, vy;                 /* position and speed */
    float px, py;                       /* previous position */
    int rotate;                         /* rotate into swimming direction */
    int mirror;                         /* going left or right? */
    int i_image;                        /* index in data file to sprite bitmap */
} ANIMAL;

ANIMAL player;                          /* actual controlled figure */

typedef struct BABY {                   /* a baby pinguin object */
    int x, y;
    int rotate, mirror;
    int i_image;                        /* index in data file to sprite bitmap */
    int origx, origy;                   /* original location */
    int distance;                       /* at what distance? */
    int number;                         /* how many'th in row? */
} BABY;

// BITMAP * txtbmpt;                    /* For debug text message */
// BITMAP * scorebmpt;                  /* For points scored */
// BITMAP * helpbmpt;                   /* For help messages */

BITMAP *playersprite;                   /* for size reference etc */
POINT mapoff;

int quit;                               /* global flag with quit request */
int help_active;                        /* help on display? */
int score;                              /* current points */
int level;                              /* current game level */
int airsupply;
#define maxair 2000
int saved_babies;
int total_babies;                       /* filled from config block */

int fish_on;                            /* do we display the fish? */
int snow_on;                            /* do we display the snow? */

int res8bit;                            /* is resolution 8 bits? */

POINT mapoff;

int debug1;                             /* displayed when debug string */
int debug2;                             /* is on display ( F12 ) */
int debug3;                             /* actual purpose may vary */

/* some auxilary routines for handling blocks etc */

void init_buffers(void);                /* allocate help buffers tec */
void exit_buffers(void);                /* clean up afterwards */

void displaylist_new(void **list, int total_items); /* get new display list */
void displaylist_draw(void *list);      /* display list */
void displaylist_clear(void **list);    /* empty list */
void displaylist_add(void **list, BITMAP *image, int x, int y, int mirror, int rotate);

BLKSTR * Collision (int x, int y);      /* simple collision routine */
void ClearCell (int x, int y);          /* remove object */
void SetCell (int x, int y, int block_id); /* set block */
BLKSTR * ConfigBlock (void);            /* get level configuration block */

void init_help(void);                   /* initialise help system */
void update_help(int level, int index); /* refresh helptext */
char *find_in_dat(int level, int index); /* find (N,M)'th block of help */
void display_helptext(int, int);        /* display help text AND overlays */
void exit_help(void);                   /* terminate help system */

void update_score(void);
void display_score(void);
void load_hiscore(void);
void save_hiscore(void);
int  test_hiscore(void);                /* return true if new hiscore */
void display_hiscore(void);

void quality_of_service(void);          /* some fancy screen dressing */
void init_snow(void);
void let_it_snow(void);
void init_fish(void);
void update_fish(void);
void draw_fish(void);
void draw_waves(void);

void display_debug_text(void);          /* generate/display debug text */
void log_state(void);                   /* for log file */
void log_system_info(void);
void check_clwp_errors(void);            /* check for tunrime clwp errors */

void init_keyhandler(void);             /* generic keyboard functions */
int read_name_flag;                     /* used in keyboard routine, */

/* some compatibility for new/old aggegro version: */

/* end of file */
