/* ------------------------------------------------------------------------
DOC <MODULE> OPTIONS
[TITLE]     Command line arguments parsing for C
[FILE]      OPTIONS.C
[VSN]       2.00
[CREATED]   931117
[LASTCHNGD] 940310
[COPYRIGHT] Copyright (C) 1993, 1994 Kees Moerman
[COMPILE]   #include "options.c"
[MODULE OF] general library module
[USES]      Borland C++ 2.00 'C' subset
[AUTHOR]    Kees Moerman
[EXTERNALS] 
[LIBRARIES] stdio stdlib strings
[TYPE]      'C' source support module
[NOTES]     Permission for use granted to PIJNENBURG PMS 940213
[PURPOSE]   This text is not displayed in the document...
This module implements a simple command line argument parser for 'C'. It 
recognised two kinf\ds of parameters: options and file names. 

Options start with either a '/' or a '-' followed by a alphabetic character 
(or a '?' for help), and end on a space or a '/' or '-'. The starting numerical
part of their value is converted to an integer.

Filenames do not start with a '/' or a '-' (but may contain those characters).
No check is made if the string contains an actual filename or allowed 
characters.

As a result of parsing, two arrays (OptionFlags and OptionPtrs) are filled
containing resp. the value of the option (if numerical) and the string
corresponing to that option. The value of these arrays can be queried
by the macros OPTION() and OPTIONPTR().

Example command line and result:
      COMLINE /a file1 file2 /B123 /caaaa/do /E3x/?
results in:
      OPTION('a') = 1     OPTIONPTR('a') = ""
      OPTION('b') = 123   OPTIONPTR('b') = "123"
      OPTION('c') = 0     OPTIONPTR('c') = "aaaa"
      OPTION('d') = 0     OPTIONPTR('d') = "o"
      OPTION('e') = 3     OPTIONPTR('e') = "3x"
      OPTION('f') = 0     OPTIONPTR('e') = NULL
      
      OptionHelp = 1
      Option_nextfilename = "file1"
      Option_nextfilename = "file2"
      Option_nextfilename = NULL

*) ENDDOC

\ --------------------------------------------------------------------------
\ Change: KM/940309  DOC extended, OptionPtrs added.
\ --------------------------------------------------------------------------

DOC OPTION()
{ char option } -- { int value }
(* the macro OPTION returns the numerical value of the option specified by 
char <option>. If not specified on the command line, this option has 
the value '0'. If 
specified without value (e.g. as in '/D') a value of 1 is returned. If 
specified a value, this value is returned.
OPTION is defined as:
      #define OPTION (OptionFlags[((x)|0x20) - 'a'])
Error Conditions:
- Parameter is not a alphabetic character
*) ENDDOC

DOC OPTIONPTR()
{ char option } -- { char *pointer }
(* the macro OPTIONPTR returns a pointer to the specified option, or NULL
if the option is not present on the command line.
OPTIONPTR is defined as:
      #define OPTIONPTR (OptionPtrs[((x)|0x20) - 'a'])
Error Conditions:
- Parameter is not a alphabetic character
*) ENDDOC

DOC Option_comline()
{ int argc, char *[]argv } -- { int result }
(* This command initialised the parameter scanner and the associated variables,
and scans the command line.
Global input variables:  OptionNumOnly,
Global output variables: OptionFlags, OptionPtrs, OptionHelp, OptionFileptr.
Error Conditions:
- None
*) ENDDOC

DOC Option_nextfilename()
{ } -- { char* filename }
(* returns (and 'eats) the next file name parameter, or NULL is not present.
Error Conditions:
- None
*) ENDDOC

DOC Option_print()
{ } -- { }
(* prints the options found on the command line, with their associated
value (both numerical and as string), and prints the filenames found
on the command line (but does not 'eat' them). Usefull for feedback from
batchfiles or during debugging. Sample output (from example in header):
Option /a = 1 ("").
Option /b = 123 ("123").
Option /c = 0 ("aaaa").
Option /d = 1 ("o").
Option /e = 3 ("3x").
Option /? = 1 ("").
Filename 'file1'
Filename 'file2'
Error Conditions:
- None
*) ENDDOC

DOC Option_process()
{ char *options } -- { }
(* processes the string options for optins (not filenames) as if specified
on the command line. Can be used to process options specified in e.g.
environment variables using the getenv() library function.
Error Conditions:
- None
*) ENDDOC

DOC OptionFlags
{ } -char[26] variable- { }
(* Output variable from Option_comline(). Global array with option values.
Error Conditions:
- None
*) ENDDOC

DOC OptionPtrs
{ } -char[26] variable- { }
(* Output variable from Option_comline(). Global array with option pointers.
Error Conditions:
- None
*) ENDDOC

DOC OptionFileptr
{ } -int variable- { }
(* Output variable from Option_comline(). This variable is used in processing 
file names on the command line. Its value is not specified, except when -1 
indicating no more file specifications are present on the command line. It is 
used by Option_nextfilename() in scanning for the next file name.
Error Conditions:
- None
*) ENDDOC

DOC OptionHelp
{ } -int variable- { }
(* Output variable from Option_comline(). True if the /? option is specified.
Error Conditions:
- None
*) ENDDOC

DOC OptionNumOnly
{ } -int variable- { }
(* Input variable to Option_comline(), defined wether non-numerical options
are allowed or not. Default value is 0: allowed. Included for compatibility 
with older software.
Error Conditions:
- None
*) ENDDOC


\ -----  End of Documentation  ------------------------------------------ */



int OptionFlags[26];                   /* global array with option values */
char *OptionPtrs[26];                  /* global array with option ptrs */
int OptionFileptr;                     /* current file ptr */
int OptionHelp;                        /* /? option? */
int OptionNumOnly;                     /* flag if numeral options only allowed */

int Option_process(char *);            /* parse options */
void Option_print(void);               /* print value of options */
int Option_comline(int, char *[]);     /* parse command line */
char *Option_nextfilename(void);       /* return next file name parameter */

#define OPTION(x)    (OptionFlags[((x)|0x20) - 'a'])
#define OPTIONPTR(x) (OptionPtrs[((x)|0x20) - 'a'])

/* End of file */
